"""Dali Gateway Scene"""

from typing import Protocol

from .helper import gen_scene_unique_id


class SupportsSceneCommands(Protocol):
    """Protocol exposing the minimum gateway interface required by Scene."""

    @property
    def gw_sn(self) -> str:
        raise NotImplementedError

    def command_write_scene(self, scene_id: int, channel: int) -> None:
        raise NotImplementedError


class Scene:
    """Dali Gateway Scene"""

    def __init__(
        self,
        command_client: SupportsSceneCommands,
        scene_id: int,
        name: str,
        channel: int,
        area_id: str,
    ) -> None:
        self._client = command_client
        self._id = scene_id
        self._name = name
        self._channel = channel
        self._area_id = area_id

    def __str__(self) -> str:
        return f"{self._name} (Channel {self._channel}, Scene {self._id})"

    def __repr__(self) -> str:
        return f"Scene(name={self._name}, unique_id={self.unique_id})"

    @property
    def unique_id(self) -> str:
        return gen_scene_unique_id(self._id, self._channel, self._client.gw_sn)

    @property
    def scene_id(self) -> int:
        return self._id

    @property
    def name(self) -> str:
        return self._name

    @property
    def gw_sn(self) -> str:
        return self._client.gw_sn

    @property
    def channel(self) -> int:
        return self._channel

    @property
    def area_id(self) -> str:
        return self._area_id

    def activate(self) -> None:
        self._client.command_write_scene(self._id, self._channel)



"""Dali Gateway Group"""

import colorsys
import logging
from typing import Any, Dict, List, Protocol, Tuple

from .helper import gen_group_unique_id

_LOGGER = logging.getLogger(__name__)


class SupportsGroupCommands(Protocol):
    """Protocol exposing the minimum gateway interface required by Group."""

    @property
    def gw_sn(self) -> str:
        raise NotImplementedError

    def command_write_group(
        self, group_id: int, channel: int, properties: List[Dict[str, Any]]
    ) -> None:
        raise NotImplementedError


class Group:
    """Dali Gateway Group"""

    def __init__(
        self,
        command_client: SupportsGroupCommands,
        group_id: int,
        name: str,
        channel: int,
        area_id: str,
    ) -> None:
        self._client = command_client
        self._id = group_id
        self._name = name
        self._channel = channel
        self._area_id = area_id

    def __str__(self) -> str:
        return f"{self._name} (Channel {self._channel}, Group {self._id})"

    def __repr__(self) -> str:
        return f"Group(name={self._name}, unique_id={self.unique_id})"

    @property
    def group_id(self) -> int:
        return self._id

    @property
    def channel(self) -> int:
        return self._channel

    @property
    def name(self) -> str:
        return self._name

    @property
    def unique_id(self) -> str:
        return gen_group_unique_id(self._id, self._channel, self._client.gw_sn)

    @property
    def gw_sn(self) -> str:
        return self._client.gw_sn

    @property
    def area_id(self) -> str:
        return self._area_id

    def _create_property(self, dpid: int, data_type: str, value: Any) -> Dict[str, Any]:
        return {"dpid": dpid, "dataType": data_type, "value": value}

    def _send_properties(self, properties: List[Dict[str, Any]]) -> None:
        for prop in properties:
            self._client.command_write_group(self._id, self._channel, [prop])

    def turn_on(
        self,
        brightness: int | None = None,
        color_temp_kelvin: int | None = None,
        rgbw_color: Tuple[float, float, float, float] | None = None,
    ) -> None:
        properties: List[Dict[str, Any]] = [self._create_property(20, "bool", True)]

        if brightness:
            properties.append(
                self._create_property(22, "uint16", brightness * 1000 / 255)
            )

        if color_temp_kelvin:
            properties.append(self._create_property(23, "uint16", color_temp_kelvin))

        if rgbw_color:
            r, g, b, w = rgbw_color
            if any([r, g, b]):
                h, s, v = colorsys.rgb_to_hsv(r / 255, g / 255, b / 255)
                h_hex = f"{int(h * 360):04x}"
                s_hex = f"{int(s * 1000):04x}"
                v_hex = f"{int(v * 1000):04x}"
                properties.append(
                    self._create_property(24, "string", f"{h_hex}{s_hex}{v_hex}")
                )

            if w > 0:
                properties.append(self._create_property(21, "uint8", int(w)))

        self._send_properties(properties)
        _LOGGER.debug(
            "Group %s (%s) turned on with properties: %s",
            self._id,
            self.name,
            properties,
        )

    def turn_off(self) -> None:
        properties: List[Dict[str, Any]] = [self._create_property(20, "bool", False)]
        self._send_properties(properties)
        _LOGGER.debug("Group %s (%s) turned off", self._id, self.name)



"""Custom exceptions for DALI Gateway operations"""


class DaliGatewayError(Exception):
    """Base exception for DALI Gateway operations"""

    def __init__(self, message: str, gw_sn: str | None = None):
        super().__init__(message)
        self.gw_sn = gw_sn



"""Test the constants for Dali Center integration."""

from PySrDaliGateway.const import DOMAIN


def test_domain_constant():
    """Test that the DOMAIN constant is correctly defined."""
    assert DOMAIN == "dali_center"
    assert isinstance(DOMAIN, str)



"""Dali Gateway Types"""

from enum import Enum
from typing import Callable, List, Tuple, TypedDict


class PanelEventType(Enum):
    """Panel button event types"""

    PRESS = "press"
    HOLD = "hold"
    DOUBLE_PRESS = "double_press"
    ROTATE = "rotate"
    RELEASE = "release"


class MotionState(Enum):
    """Motion sensor state types"""

    NO_MOTION = "no_motion"
    MOTION = "motion"
    VACANT = "vacant"
    OCCUPANCY = "occupancy"
    PRESENCE = "presence"


class DeviceProperty:
    dpid: int
    data_type: str


class SceneDeviceProperty(TypedDict):
    dpid: int
    data_type: str
    value: int


class LightStatus(TypedDict):
    """Status for lighting devices (Dimmer, CCT, RGB, RGBW, RGBWA)"""

    is_on: bool | None
    brightness: int | None  # 0-255
    color_temp_kelvin: int | None
    hs_color: Tuple[float, float] | None  # hue (0-360), saturation (0-100)
    rgbw_color: Tuple[int, int, int, int] | None  # r,g,b,w (0-255 each)
    white_level: int | None  # 0-255


class SceneDeviceType(TypedDict):
    dev_type: str
    channel: int
    address: int
    gw_sn_obj: str
    property: LightStatus


class VersionType(TypedDict):
    software: str
    firmware: str


class DeviceParamType(TypedDict):
    # address: int
    # fade_time: int
    # fade_rate: int
    # power_status: int
    # system_failure_status: int
    max_brightness: int
    # min_brightness: int
    # standby_power: int
    # max_power: int
    # cct_cool: int
    # cct_warm: int
    # phy_cct_cool: int
    # phy_cct_warm: int
    # step_cct: int
    # temp_thresholds: int
    # runtime_thresholds: int


class PanelConfig(TypedDict):
    """Panel configuration type definition."""

    button_count: int
    events: List[str]


class PanelStatus(TypedDict):
    """Status for control panels (2-Key, 4-Key, 6-Key, 8-Key)"""

    event_name: str  # button_{key_no}_{event_type}
    key_no: int  # Button number
    event_type: PanelEventType  # press, hold, double_press, rotate, release
    rotate_value: int | None  # For rotate events (only for rotate event type)


class MotionStatus(TypedDict):
    """Status for motion sensor devices"""

    motion_state: MotionState
    dpid: int  # The original dpid that triggered this state


class IlluminanceStatus(TypedDict):
    """Status for illuminance sensor devices"""

    illuminance_value: float  # Illuminance in lux
    is_valid: bool  # Whether the value is within valid range (0-1000)


LightStatusCallback = Callable[[LightStatus], None]
PanelStatusCallback = Callable[[PanelStatus], None]
MotionStatusCallback = Callable[[MotionStatus], None]
IlluminanceStatusCallback = Callable[[IlluminanceStatus], None]



"""Constants for the Dali Center."""

from importlib import resources

from .types import PanelConfig

DOMAIN = "dali_center"

DEVICE_MODEL_MAP = {
    "0101": "DALI DT6 Dimmable Driver",
    "0102": "DALI DT8 Tc Dimmable Driver",
    "0103": "DALI DT8 RGB Dimmable Driver",
    "0104": "DALI DT8 XY Dimmable Driver",
    "0105": "DALI DT8 RGBW Dimmable Driver",
    "0106": "DALI DT8 RGBWA Dimmable Driver",
    "0201": "DALI-2 Motion Sensor",
    "020101": "DALI-2 Motion Sensor",
    "020102": "DALI-2 Motion Sensor",
    "020103": "DALI-2 Motion Sensor",
    "020104": "DALI-2 Motion Sensor",
    "020105": "DALI-2 Motion Sensor",
    "020106": "DALI-2 Motion Sensor",
    "020107": "DALI-2 Motion Sensor",
    "020108": "DALI-2 Motion Sensor",
    "020109": "DALI-2 Motion Sensor",
    "020110": "DALI-2 Motion Sensor",
    "020111": "DALI-2 Motion Sensor",
    "020112": "DALI-2 Motion Sensor",
    "020113": "DALI-2 Motion Sensor",
    "020114": "DALI-2 Motion Sensor",
    "020115": "DALI-2 Motion Sensor",
    "020116": "DALI-2 Motion Sensor",
    "020117": "DALI-2 Motion Sensor",
    "020118": "DALI-2 Motion Sensor",
    "020119": "DALI-2 Motion Sensor",
    "020120": "DALI-2 Motion Sensor",
    "0202": "DALI-2 Illuminance Sensor",
    "0302": "DALI-2 2-Key Push Button Panel",
    "0304": "DALI-2 4-Key Push Button Panel",
    "0306": "DALI-2 6-Key Push Button Panel",
    "0308": "DALI-2 8-Key Push Button Panel",
}

DEVICE_TYPE_MAP = {
    "0101": "Dimmer",
    "0102": "CCT",
    "0103": "RGB",
    "0104": "XY",
    "0105": "RGBW",
    "0106": "RGBWA",
    "0201": "Motion",
    "020101": "Motion (1)",
    "020102": "Motion (2)",
    "020103": "Motion (3)",
    "020104": "Motion (4)",
    "020105": "Motion (5)",
    "020106": "Motion (6)",
    "020107": "Motion (7)",
    "020108": "Motion (8)",
    "020109": "Motion (9)",
    "020110": "Motion (10)",
    "020111": "Motion (11)",
    "020112": "Motion (12)",
    "020113": "Motion (13)",
    "020114": "Motion (14)",
    "020115": "Motion (15)",
    "020116": "Motion (16)",
    "020117": "Motion (17)",
    "020118": "Motion (18)",
    "020119": "Motion (19)",
    "020120": "Motion (20)",
    "0202": "Illuminance",
    "0302": "2-Key Panel",
    "0304": "4-Key Panel",
    "0306": "6-Key Panel",
    "0308": "8-Key Panel",
}

COLOR_MODE_MAP = {
    "0102": "color_temp",  # CCT
    "0103": "hs",  # RGB
    "0104": "hs",  # XY
    "0105": "rgbw",  # RGBW
    "0106": "rgbw",  # RGBWA
}

BUTTON_EVENTS = {
    1: "press",
    2: "hold",
    3: "double_press",
    4: "rotate",
    5: "release",
}

PANEL_CONFIGS: dict[str, PanelConfig] = {
    "0302": {  # 2-button panel
        "button_count": 2,
        "events": ["press", "hold", "double_press", "release"],
    },
    "0304": {  # 4-button panel
        "button_count": 4,
        "events": ["press", "hold", "double_press", "release"],
    },
    "0306": {  # 6-button panel
        "button_count": 6,
        "events": ["press", "hold", "double_press", "release"],
    },
    "0308": {  # 8-button panel
        "button_count": 8,
        "events": ["press", "hold", "double_press", "release"],
    },
    "0300": {  # rotary knob panel
        "button_count": 1,
        "events": ["press", "double_press", "rotate"],
    },
}

CA_CERT_PATH = resources.files("PySrDaliGateway") / "certs" / "ca.crt"



"""Dali Gateway Discovery"""

import asyncio
import contextlib
import ipaddress
import json
import logging
import socket
from typing import Any, Dict, List, Set
import uuid

from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
import psutil

from .gateway import DaliGateway

_LOGGER = logging.getLogger(__name__)


class NetworkManager:
    """Network interface manager"""

    def get_valid_interfaces(self) -> List[Dict[str, Any]]:
        interfaces: List[Dict[str, Any]] = []
        _LOGGER.debug("Scanning network interfaces for discovery")
        for interface_name, addrs in psutil.net_if_addrs().items():
            for addr in addrs:
                if addr.family == socket.AF_INET:
                    ip = addr.address
                    if self.is_valid_ip(ip):
                        interface_info = self.create_interface_info(interface_name, ip)
                        interfaces.append(interface_info)
                        _LOGGER.debug(
                            "Found valid interface: %s (%s)", interface_name, ip
                        )
                    else:
                        _LOGGER.debug(
                            "Skipping invalid IP: %s on %s", ip, interface_name
                        )

        _LOGGER.debug(
            "Network scan complete. %d valid interfaces found", len(interfaces)
        )
        return interfaces

    def is_valid_ip(self, ip: str) -> bool:
        if not ip or ip.startswith("127."):
            return False
        ip_obj = ipaddress.IPv4Address(ip)
        return not ip_obj.is_loopback and not ip_obj.is_link_local

    def create_interface_info(self, name: str, ip: str) -> Dict[str, Any]:
        return {"name": name, "address": ip, "network": f"{ip}/24"}


class MessageCryptor:
    """Message encryption and decryption handler"""

    SR_KEY: str = "SR-DALI-GW-HASYS"
    ENCRYPTION_IV: bytes = b"0000000000101111"

    def encrypt_data(self, data: str, key: str) -> str:
        key_bytes = key.encode("utf-8")
        cipher = Cipher(algorithms.AES(key_bytes), modes.CTR(self.ENCRYPTION_IV))
        encryptor = cipher.encryptor()
        encrypted_data = encryptor.update(data.encode("utf-8")) + encryptor.finalize()
        return encrypted_data.hex()

    def decrypt_data(self, encrypted_hex: str, key: str) -> str:
        key_bytes = key.encode("utf-8")
        encrypted_bytes = bytes.fromhex(encrypted_hex)
        cipher = Cipher(algorithms.AES(key_bytes), modes.CTR(self.ENCRYPTION_IV))
        decryptor = cipher.decryptor()
        decrypted_data = decryptor.update(encrypted_bytes) + decryptor.finalize()
        return decrypted_data.decode("utf-8")

    def random_key(self) -> str:
        return uuid.uuid4().hex[:16]

    def prepare_discovery_message(self, gw_sn: str | None = None) -> bytes:
        key = self.random_key()
        msg_enc = self.encrypt_data("discover", key)
        combined_data = key + msg_enc
        cmd = self.encrypt_data(combined_data, self.SR_KEY)
        message_dict = {"cmd": cmd, "type": "HA", "snList": []}
        if gw_sn:
            message_dict["snList"] = [gw_sn]

        _LOGGER.debug("Prepared discovery message: %s", message_dict)
        message_json = json.dumps(message_dict)
        return message_json.encode("utf-8")


class MulticastSender:
    """Multicast communication manager"""

    MULTICAST_ADDR: str = "239.255.255.250"
    SEND_PORT: int = 1900
    LISTEN_PORT: int = 50569

    def create_listener_socket(self, interfaces: List[Dict[str, Any]]) -> socket.socket:
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        with contextlib.suppress(AttributeError, OSError):
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        self._bind_to_port(sock)
        self._join_multicast_groups(sock, interfaces)
        sock.setblocking(False)
        return sock

    def cleanup_socket(
        self, sock: socket.socket, interfaces: List[Dict[str, Any]]
    ) -> None:
        mreq_list = []
        for interface in interfaces:
            mreq = socket.inet_aton(self.MULTICAST_ADDR) + socket.inet_aton(
                interface["address"]
            )
            mreq_list.append(mreq)
        for mreq in mreq_list:
            with contextlib.suppress(OSError):
                sock.setsockopt(socket.IPPROTO_IP, socket.IP_DROP_MEMBERSHIP, mreq)
        sock.close()

    async def send_multicast_message(
        self, interfaces: List[Dict[str, Any]], message: bytes
    ) -> None:
        tasks = [
            asyncio.create_task(self._send_on_interface(interface, message))
            for interface in interfaces
        ]
        await asyncio.gather(*tasks, return_exceptions=True)

    def _bind_to_port(self, sock: socket.socket) -> None:
        ports_to_try = [
            self.LISTEN_PORT,
            *range(self.LISTEN_PORT + 1, self.LISTEN_PORT + 10),
            0,
        ]

        def try_bind_port(port: int) -> bool:
            try:
                sock.bind(("0.0.0.0", port))
            except OSError:
                _LOGGER.debug("Port %d unavailable, trying next port", port)
                return False
            else:
                _LOGGER.debug("Successfully bound listener socket to port %d", port)
                return True

        for port in ports_to_try:
            if try_bind_port(port):
                return

        _LOGGER.error("Unable to bind to any port after trying all options")
        raise OSError("Unable to bind to any port")

    def _join_multicast_groups(
        self, sock: socket.socket, interfaces: List[Dict[str, Any]]
    ) -> None:
        _LOGGER.debug("Joining multicast groups on %d interfaces", len(interfaces))

        for interface in interfaces:
            mreq = socket.inet_aton(self.MULTICAST_ADDR) + socket.inet_aton(
                interface["address"]
            )
            with contextlib.suppress(OSError):
                sock.setsockopt(socket.IPPROTO_IP, socket.IP_DROP_MEMBERSHIP, mreq)

            sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
            _LOGGER.debug(
                "Joined multicast group on interface %s (%s)",
                interface["name"],
                interface["address"],
            )

    async def _send_on_interface(
        self, interface: Dict[str, Any], message: bytes
    ) -> None:
        with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as sock:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
            sock.bind((interface["address"], 0))
            sock.setsockopt(
                socket.IPPROTO_IP,
                socket.IP_MULTICAST_IF,
                socket.inet_aton(interface["address"]),
            )
            sock.sendto(message, (self.MULTICAST_ADDR, self.SEND_PORT))


class DaliGatewayDiscovery:
    """Dali Gateway Discovery"""

    DISCOVERY_TIMEOUT: float = 180.0
    SEND_INTERVAL: float = 2.0

    def __init__(self) -> None:
        self.network_manager = NetworkManager()
        self.cryptor = MessageCryptor()
        self.sender = MulticastSender()

    async def discover_gateways(self, gw_sn: str | None = None) -> List[DaliGateway]:
        _LOGGER.info(
            "Starting DALI gateway discovery%s",
            f" for specific gateway: {gw_sn}" if gw_sn else "",
        )

        interfaces = self.network_manager.get_valid_interfaces()
        _LOGGER.debug(
            "Found %d valid network interfaces: %s",
            len(interfaces),
            [iface["name"] for iface in interfaces],
        )

        if not interfaces:
            _LOGGER.error(
                "No valid network interfaces found for discovery - check network connection"
            )
            return []

        message = self.cryptor.prepare_discovery_message(gw_sn)
        listen_sock = self.sender.create_listener_socket(interfaces)

        try:
            gateways = await self._run_discovery(listen_sock, interfaces, message)
            _LOGGER.info("Discovery completed. Found %d gateway(s)", len(gateways))
            return gateways
        finally:
            self.sender.cleanup_socket(listen_sock, interfaces)

    async def _run_discovery(
        self, sock: socket.socket, interfaces: List[Dict[str, Any]], message: bytes
    ) -> List[DaliGateway]:
        start_time = asyncio.get_event_loop().time()
        first_gateway_found = asyncio.Event()
        unique_gateways: List[DaliGateway] = []
        seen_sns: Set[str] = set()

        # Sender task
        sender_task = asyncio.create_task(
            self._sender_loop(interfaces, message, first_gateway_found, start_time)
        )

        # Receiver task
        receiver_task = asyncio.create_task(
            self._receiver_loop(
                sock, first_gateway_found, start_time, unique_gateways, seen_sns
            )
        )

        _, pending = await asyncio.wait(
            [sender_task, receiver_task], return_when=asyncio.FIRST_COMPLETED
        )
        for task in pending:
            task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await task
        return unique_gateways

    async def _sender_loop(
        self,
        interfaces: List[Dict[str, Any]],
        message: bytes,
        first_gateway_found: asyncio.Event,
        start_time: float,
    ) -> None:
        send_count = 0
        while not first_gateway_found.is_set():
            if asyncio.get_event_loop().time() - start_time >= self.DISCOVERY_TIMEOUT:
                _LOGGER.info(
                    "Discovery timeout reached after %.1f seconds",
                    self.DISCOVERY_TIMEOUT,
                )
                break

            send_count += 1
            _LOGGER.debug(
                "Sending discovery message #%d to %d interfaces",
                send_count,
                len(interfaces),
            )
            await self.sender.send_multicast_message(interfaces, message)

            try:
                await asyncio.wait_for(
                    first_gateway_found.wait(), timeout=self.SEND_INTERVAL
                )
                _LOGGER.debug(
                    "Gateway found, stopping sender loop after %d sends", send_count
                )
                break
            except asyncio.TimeoutError:
                continue

    async def _receiver_loop(
        self,
        sock: socket.socket,
        first_gateway_found: asyncio.Event,
        start_time: float,
        unique_gateways: List[DaliGateway],
        seen_sns: Set[str],
    ) -> None:
        _LOGGER.debug("Starting receiver loop, listening on socket")

        while not first_gateway_found.is_set():
            if asyncio.get_event_loop().time() - start_time >= self.DISCOVERY_TIMEOUT:
                _LOGGER.debug("Receiver loop timeout reached")
                break

            addr = None
            try:
                await asyncio.sleep(0.1)
                data, addr = sock.recvfrom(1024)
                _LOGGER.debug("Received response from %s, processing data", addr)

                response_json = json.loads(data.decode("utf-8"))
                raw_data = response_json.get("data")

                if raw_data and raw_data.get("gwSn") not in seen_sns:
                    if gateway := self._process_gateway_data(raw_data):
                        _LOGGER.info(
                            "Discovered gateway: %s (%s) at %s:%s",
                            gateway.name,
                            gateway.gw_sn,
                            gateway.gw_ip,
                            gateway.port,
                        )
                        unique_gateways.append(gateway)
                        seen_sns.add(gateway.gw_sn)
                        first_gateway_found.set()
                        break
                    _LOGGER.warning("Failed to process gateway data from %s", addr)
                elif raw_data and raw_data.get("gwSn") in seen_sns:
                    _LOGGER.debug(
                        "Ignoring duplicate gateway response: %s", raw_data.get("gwSn")
                    )

            except json.JSONDecodeError as exc:
                _LOGGER.warning(
                    "Invalid JSON response from %s: %s. Raw data: %s",
                    addr or "unknown",
                    exc,
                    data[:100] if data else "<empty>",
                )
                continue
            except (BlockingIOError, asyncio.CancelledError):
                continue
            except OSError as exc:
                _LOGGER.error(
                    "Socket error in receiver loop - this may prevent discovery: %s",
                    exc,
                )
                break

    def _process_gateway_data(self, raw_data: Any) -> DaliGateway | None:
        gw_sn = raw_data.get("gwSn")
        if not gw_sn:
            _LOGGER.warning("Gateway data missing required 'gwSn' field")
            return None

        _LOGGER.debug("Processing gateway data for: %s", gw_sn)

        encrypted_user = raw_data.get("username", "")
        encrypted_pass = raw_data.get("passwd", "")

        try:
            decrypted_user = self.cryptor.decrypt_data(
                encrypted_user, self.cryptor.SR_KEY
            )
            decrypted_pass = self.cryptor.decrypt_data(
                encrypted_pass, self.cryptor.SR_KEY
            )
        except UnicodeDecodeError as e:
            _LOGGER.error(
                "Failed to decrypt gateway credentials for %s: %s. "
                "This gateway will be skipped.",
                gw_sn,
                e,
            )
            return None

        gateway_name = raw_data.get("name") or f"Dali Gateway {gw_sn}"
        channel_total = [
            int(ch)
            for ch in raw_data.get("channelTotal", [])
            if isinstance(ch, (int, str)) and str(ch).isdigit()
        ]

        gateway = DaliGateway(
            gw_sn=gw_sn,
            gw_ip=raw_data.get("gwIp"),
            port=int(raw_data.get("port", 0)),
            username=decrypted_user,
            passwd=decrypted_pass,
            name=gateway_name,
            channel_total=channel_total,
            is_tls=bool(raw_data.get("isMqttTls")),
        )

        _LOGGER.debug("Successfully processed gateway: %s", gateway_name)
        return gateway



"""Dali Gateway Device"""

import colorsys
import logging
from typing import Any, Dict, Iterable, List, Protocol, Tuple

from .const import COLOR_MODE_MAP
from .types import DeviceProperty

_LOGGER = logging.getLogger(__name__)


class SupportsDeviceCommands(Protocol):
    """Protocol exposing the gateway commands needed by Device instances."""

    @property
    def gw_sn(self) -> str:
        raise NotImplementedError

    def command_write_dev(
        self,
        dev_type: str,
        channel: int,
        address: int,
        properties: List[Dict[str, Any]],
    ) -> None:
        raise NotImplementedError

    def command_read_dev(self, dev_type: str, channel: int, address: int) -> None:
        raise NotImplementedError

    def command_get_energy(
        self, dev_type: str, channel: int, address: int, year: int, month: int, day: int
    ) -> None:
        raise NotImplementedError

    def command_set_sensor_on_off(
        self, dev_type: str, channel: int, address: int, value: bool
    ) -> None:
        raise NotImplementedError

    def command_get_sensor_on_off(
        self, dev_type: str, channel: int, address: int
    ) -> None:
        raise NotImplementedError


class Device:
    """Dali Gateway Device"""

    def __init__(
        self,
        command_client: SupportsDeviceCommands,
        unique_id: str,
        dev_id: str,
        name: str,
        dev_type: str,
        channel: int,
        address: int,
        status: str,
        *,
        dev_sn: str,
        area_name: str,
        area_id: str,
        model: str,
        properties: Iterable[DeviceProperty] | None = None,
    ) -> None:
        self._client = command_client
        self._unique_id = unique_id
        self._id = dev_id
        self._name = name
        self._dev_type = dev_type
        self._channel = channel
        self._address = address
        self._status = status
        self._dev_sn = dev_sn
        self._area_name = area_name
        self._area_id = area_id
        self._model = model
        self._properties: List[DeviceProperty] = list(properties or [])

    def __repr__(self) -> str:
        return f"Device(name={self._name}, unique_id={self.unique_id})"

    def __str__(self) -> str:
        return self._name

    @property
    def gw_sn(self) -> str:
        return self._client.gw_sn

    @property
    def dev_id(self) -> str:
        return self._id

    @property
    def dev_type(self) -> str:
        return self._dev_type

    @property
    def status(self) -> str:
        return self._status

    @property
    def name(self) -> str:
        return self._name

    @property
    def unique_id(self) -> str:
        return self._unique_id

    @property
    def channel(self) -> int:
        return self._channel

    @property
    def address(self) -> int:
        return self._address

    @property
    def color_mode(self) -> str:
        return COLOR_MODE_MAP.get(self._dev_type, "brightness")

    @property
    def model(self) -> str:
        return self._model

    @property
    def area_name(self) -> str:
        return self._area_name

    @property
    def area_id(self) -> str:
        return self._area_id

    @property
    def dev_sn(self) -> str:
        return self._dev_sn

    @property
    def properties(self) -> List[DeviceProperty]:
        return list(self._properties)

    def _create_property(self, dpid: int, data_type: str, value: Any) -> Dict[str, Any]:
        return {"dpid": dpid, "dataType": data_type, "value": value}

    def _send_properties(self, properties: List[Dict[str, Any]]) -> None:
        for prop in properties:
            self._client.command_write_dev(
                self._dev_type, self._channel, self._address, [prop]
            )

    def turn_on(
        self,
        brightness: int | None = None,
        color_temp_kelvin: int | None = None,
        hs_color: Tuple[float, float] | None = None,
        rgbw_color: Tuple[float, float, float, float] | None = None,
    ) -> None:
        properties = [self._create_property(20, "bool", True)]

        if brightness:
            properties.append(
                self._create_property(22, "uint16", brightness * 1000 / 255)
            )

        if color_temp_kelvin:
            properties.append(self._create_property(23, "uint16", color_temp_kelvin))

        if hs_color:
            h, s = hs_color
            h_hex = f"{int(h):04x}"
            s_hex = f"{int(s * 1000 / 100):04x}"
            v_hex = f"{1000:04x}"
            properties.append(
                self._create_property(24, "string", f"{h_hex}{s_hex}{v_hex}")
            )

        if rgbw_color:
            r, g, b, w = rgbw_color
            if any([r, g, b]):
                h, s, v = colorsys.rgb_to_hsv(r / 255, g / 255, b / 255)
                h_hex = f"{int(h * 360):04x}"
                s_hex = f"{int(s * 1000):04x}"
                v_hex = f"{int(v * 1000):04x}"
                properties.append(
                    self._create_property(24, "string", f"{h_hex}{s_hex}{v_hex}")
                )

            if w > 0:
                properties.append(self._create_property(21, "uint8", int(w)))

        self._send_properties(properties)
        _LOGGER.debug(
            "Device %s (%s) turned on with properties: %s",
            self._id,
            self._name,
            properties,
        )

    def turn_off(self) -> None:
        properties = [self._create_property(20, "bool", False)]
        self._send_properties(properties)
        _LOGGER.debug("Device %s (%s) turned off", self._id, self._name)

    def read_status(self) -> None:
        self._client.command_read_dev(self._dev_type, self._channel, self._address)
        _LOGGER.debug("Requesting status for device %s (%s)", self.dev_id, self._name)

    def press_button(self, button_id: int, event_type: int = 1) -> None:
        properties = [self._create_property(button_id, "uint8", event_type)]

        self._send_properties(properties)
        _LOGGER.debug(
            "Button %d pressed on device %s (%s) with event type %d",
            button_id,
            self._id,
            self._name,
            event_type,
        )

    def set_sensor_enabled(self, enabled: bool) -> None:
        self._client.command_set_sensor_on_off(
            self._dev_type, self._channel, self._address, enabled
        )
        _LOGGER.debug(
            "Sensor %s (%s) enabled state set to %s",
            self._id,
            self._name,
            enabled,
        )

    def get_sensor_enabled(self) -> None:
        self._client.command_get_sensor_on_off(
            self._dev_type, self._channel, self._address
        )
        _LOGGER.debug(
            "Requesting sensor %s (%s) enabled state",
            self._id,
            self._name,
        )

    def get_energy(self, year: int, month: int, day: int) -> None:
        self._client.command_get_energy(
            self._dev_type,
            self._channel,
            self._address,
            year,
            month,
            day,
        )
        _LOGGER.debug(
            "Requesting energy data for device %s (%s)",
            self._id,
            self._name,
        )



"""Helper functions for Dali Gateway"""

import colorsys
from typing import Any, Dict, List

from .const import BUTTON_EVENTS, DEVICE_TYPE_MAP
from .types import (
    IlluminanceStatus,
    LightStatus,
    MotionState,
    MotionStatus,
    PanelEventType,
    PanelStatus,
)


def is_light_device(dev_type: str) -> bool:
    return dev_type.startswith("01")


def is_motion_sensor(dev_type: str) -> bool:
    return dev_type.startswith("0201")


def is_illuminance_sensor(dev_type: str) -> bool:
    return dev_type == "0202"


def is_panel_device(dev_type: str) -> bool:
    return dev_type.startswith("03")


def is_sensor_device(dev_type: str) -> bool:
    return is_motion_sensor(dev_type) or is_illuminance_sensor(dev_type)


def gen_device_unique_id(dev_type: str, channel: int, address: int, gw_sn: str) -> str:
    return f"{dev_type}{channel:04d}{address:02d}{gw_sn}"


def gen_device_name(dev_type: str, channel: int, address: int) -> str:
    if dev_type in DEVICE_TYPE_MAP:
        type_name = DEVICE_TYPE_MAP[dev_type]
        return f"{type_name} {channel:04d}-{address:02d}"
    if dev_type:
        return f"Device {dev_type} {channel:04d}-{address:02d}"
    raise ValueError(f"Invalid device type: {dev_type}")


def gen_group_unique_id(group_id: int, channel: int, gw_sn: str) -> str:
    return f"group_{group_id:04d}_{channel:04d}_{gw_sn}"


def gen_scene_unique_id(scene_id: int, channel: int, gw_sn: str) -> str:
    return f"scene_{scene_id:04d}_{channel:04d}_{gw_sn}"


def parse_light_status(property_list: List[Dict[str, Any]]) -> LightStatus:
    """Parse raw property list into LightStatus object for light devices"""

    props: Dict[int, Any] = {}
    for prop in property_list:
        prop_id = prop.get("id") or prop.get("dpid")
        value = prop.get("value")
        if prop_id is not None and value is not None:
            props[prop_id] = value

    light_status: LightStatus = {
        "is_on": None,
        "brightness": None,
        "color_temp_kelvin": None,
        "hs_color": None,
        "rgbw_color": None,
        "white_level": None,
    }

    if 20 in props:
        light_status["is_on"] = bool(props[20])

    if 21 in props:
        white_level = int(props[21])
        light_status["white_level"] = min(255, max(0, white_level))
        if light_status["rgbw_color"] is not None:
            r, g, b, _ = light_status["rgbw_color"]
            light_status["rgbw_color"] = (r, g, b, white_level)

    if 22 in props:
        brightness_value = float(props[22])
        if brightness_value == 0 and light_status["brightness"] is None:
            light_status["brightness"] = 255
        else:
            light_status["brightness"] = int(brightness_value / 1000 * 255)

    if 23 in props:
        light_status["color_temp_kelvin"] = int(props[23])

    if 24 in props:
        hsv_str = str(props[24])

        h = int(hsv_str[0:4], 16)
        s = int(hsv_str[4:8], 16)

        if len(hsv_str) == 8:
            s_percentage = s / 10
            light_status["hs_color"] = (float(h), float(s_percentage))

        elif len(hsv_str) >= 12:
            v = int(hsv_str[8:12], 16)

            h_norm = max(0, min(360, h)) / 360.0
            s_norm = max(0, min(1000, s)) / 1000.0
            v_norm = max(0, min(1000, v)) / 1000.0

            if v_norm == 0:
                v_norm = 1

            rgb = colorsys.hsv_to_rgb(h_norm, s_norm, v_norm)
            w = (
                light_status["white_level"]
                if light_status["white_level"] is not None
                else 0
            )

            light_status["rgbw_color"] = (
                int(rgb[0] * 255),
                int(rgb[1] * 255),
                int(rgb[2] * 255),
                w,
            )

    return light_status


def parse_panel_status(property_list: List[Dict[str, Any]]) -> List[PanelStatus]:
    """Parse raw property list into PanelStatus objects for panel devices"""

    panel_events: List[PanelStatus] = []

    for prop in property_list:
        dpid = prop.get("dpid")
        key_no = prop.get("keyNo")
        value = prop.get("value")

        if dpid is None or key_no is None:
            continue

        event_type_str = BUTTON_EVENTS.get(dpid)
        if not event_type_str:
            continue

        # Convert string to enum
        try:
            event_type = PanelEventType(event_type_str)
        except ValueError:
            continue

        event_name = f"button_{key_no}_{event_type_str}"

        panel_status: PanelStatus = {
            "event_name": event_name,
            "key_no": key_no,
            "event_type": event_type,
            "rotate_value": value if event_type == PanelEventType.ROTATE else None,
        }

        panel_events.append(panel_status)

    return panel_events


def parse_motion_status(property_list: List[Dict[str, Any]]) -> List[MotionStatus]:
    """Parse raw property list into MotionStatus objects for motion sensor devices"""

    motion_events: List[MotionStatus] = []

    # Motion state mapping based on dpid values
    motion_map = {
        1: MotionState.NO_MOTION,
        2: MotionState.MOTION,
        3: MotionState.VACANT,
        4: MotionState.OCCUPANCY,
        5: MotionState.PRESENCE,
    }

    for prop in property_list:
        dpid = prop.get("dpid")

        if dpid is None:
            continue

        motion_state = motion_map.get(dpid)
        if motion_state is None:
            # Default to no_motion for unknown dpid values
            motion_state = MotionState.NO_MOTION

        motion_status: MotionStatus = {
            "motion_state": motion_state,
            "dpid": dpid,
        }

        motion_events.append(motion_status)

    return motion_events


def parse_illuminance_status(
    property_list: List[Dict[str, Any]],
) -> List[IlluminanceStatus]:
    """Parse raw property list into IlluminanceStatus objects for illuminance sensor devices"""

    illuminance_events: List[IlluminanceStatus] = []

    for prop in property_list:
        dpid = prop.get("dpid")
        value = prop.get("value")

        # Handle illuminance sensor status (dpid 4 for illuminance value)
        if dpid == 4 and value is not None:
            try:
                illuminance_value = float(value)
                # Check if value is within valid range
                is_valid = not (illuminance_value > 1000 or illuminance_value <= 0)

                illuminance_status: IlluminanceStatus = {
                    "illuminance_value": illuminance_value,
                    "is_valid": is_valid,
                }

                illuminance_events.append(illuminance_status)
            except (ValueError, TypeError):
                # Skip invalid values
                continue

    return illuminance_events



"""Dali Gateway Panel Device"""

from typing import Iterable, List

from .const import PANEL_CONFIGS
from .device import Device, SupportsDeviceCommands
from .types import DeviceProperty


class Panel(Device):
    """Dali Gateway Panel Device"""

    def __init__(
        self,
        command_client: SupportsDeviceCommands,
        unique_id: str,
        dev_id: str,
        name: str,
        dev_type: str,
        channel: int,
        address: int,
        status: str,
        *,
        dev_sn: str,
        area_name: str,
        area_id: str,
        model: str,
        properties: Iterable[DeviceProperty] | None = None,
    ) -> None:
        super().__init__(
            command_client,
            unique_id,
            dev_id,
            name,
            dev_type,
            channel,
            address,
            status,
            dev_sn=dev_sn,
            area_name=area_name,
            area_id=area_id,
            model=model,
            properties=properties,
        )
        self._panel_config = PANEL_CONFIGS.get(
            self._dev_type, {"button_count": 1, "events": ["press"]}
        )

    @property
    def button_count(self) -> int:
        """Get the number of buttons on this panel"""
        return self._panel_config.get("button_count", 1)

    @property
    def supported_events(self) -> List[str]:
        """Get supported events for this panel type"""
        return self._panel_config.get("events", ["press"])

    def get_available_event_types(self) -> List[str]:
        """Generate all possible event types for this panel device"""
        event_types: List[str] = []
        event_types.extend(
            f"button_{button_num}_{event}"
            for button_num in range(1, self.button_count + 1)
            for event in self.supported_events
        )
        return event_types



"""Dali Gateway"""
# pylint: disable=invalid-name

from .__version__ import __version__
from .device import Device
from .gateway import DaliGateway
from .group import Group
from .panel import Panel
from .scene import Scene
from .types import (
    IlluminanceStatus,
    LightStatus,
    MotionState,
    MotionStatus,
    PanelEventType,
    PanelStatus,
    VersionType,
)

__all__ = [
    "DaliGateway",
    "Device",
    "Group",
    "IlluminanceStatus",
    "LightStatus",
    "MotionState",
    "MotionStatus",
    "Panel",
    "PanelEventType",
    "PanelStatus",
    "Scene",
    "VersionType",
    "__version__",
]



"""Dali Gateway"""

import asyncio
import json
import logging
import ssl
import time
from typing import Any, Callable, Dict, List, Sequence

import paho.mqtt.client as paho_mqtt

# Backward compatibility with paho-mqtt < 2.0.0
try:
    from paho.mqtt.enums import CallbackAPIVersion

    HAS_CALLBACK_API_VERSION = True
except ImportError:
    # paho-mqtt < 2.0.0 doesn't have CallbackAPIVersion
    HAS_CALLBACK_API_VERSION = False  # pyright: ignore[reportConstantRedefinition]

from .const import CA_CERT_PATH, DEVICE_MODEL_MAP
from .device import Device
from .exceptions import DaliGatewayError
from .group import Group
from .helper import (
    gen_device_name,
    gen_device_unique_id,
    gen_group_unique_id,
    gen_scene_unique_id,
    is_illuminance_sensor,
    is_light_device,
    is_motion_sensor,
    is_panel_device,
    parse_illuminance_status,
    parse_light_status,
    parse_motion_status,
    parse_panel_status,
)
from .scene import Scene
from .types import (
    DeviceParamType,
    IlluminanceStatus,
    LightStatus,
    MotionStatus,
    PanelStatus,
    SceneDeviceType,
    VersionType,
)

_LOGGER = logging.getLogger(__name__)


class DaliGateway:
    """Dali Gateway"""

    def __init__(
        self,
        gw_sn: str,
        gw_ip: str,
        port: int,
        username: str,
        passwd: str,
        *,
        name: str | None = None,
        channel_total: Sequence[int] | None = None,
        is_tls: bool = False,
    ) -> None:
        # Gateway information
        self._gw_sn = gw_sn
        self._gw_ip = gw_ip
        self._port = port
        self._name = name or gw_sn
        self._username = username
        self._passwd = passwd
        self._is_tls = is_tls
        self._channel_total = (
            [int(ch) for ch in channel_total] if channel_total else [0]
        )

        # MQTT topics
        self._sub_topic = f"/{self._gw_sn}/client/reciver/"
        self._pub_topic = f"/{self._gw_sn}/server/publish/"

        # MQTT client - handle compatibility between paho-mqtt versions
        if HAS_CALLBACK_API_VERSION:
            # paho-mqtt >= 2.0.0
            self._mqtt_client = paho_mqtt.Client(
                CallbackAPIVersion.VERSION2,  # pyright: ignore[reportPossiblyUnboundVariable]
                client_id=f"ha_dali_center_{self._gw_sn}",
                protocol=paho_mqtt.MQTTv311,
            )
        else:
            # paho-mqtt < 2.0.0
            self._mqtt_client = paho_mqtt.Client(
                client_id=f"ha_dali_center_{self._gw_sn}",
                protocol=paho_mqtt.MQTTv311,
            )

        self._mqtt_client.enable_logger()

        # Connection result
        self._connect_result: int | None = None
        self._connection_event = asyncio.Event()

        # Set up client callbacks
        self._mqtt_client.on_connect = self._on_connect
        self._mqtt_client.on_disconnect = self._on_disconnect
        self._mqtt_client.on_message = self._on_message

        # Scene/Group/Device Received
        self._scenes_received = asyncio.Event()
        self._groups_received = asyncio.Event()
        self._devices_received = asyncio.Event()
        self._version_received = asyncio.Event()

        self._scenes_result: list[Scene] = []
        self._groups_result: list[Group] = []
        self._devices_result: list[Device] = []
        self._version_result: VersionType | None = None
        self._read_group_received = asyncio.Event()
        self._read_group_result: Dict[str, Any] | None = None
        self._read_scene_received = asyncio.Event()
        self._read_scene_result: Dict[str, Any] | None = None

        # Callbacks
        self._on_online_status: Callable[[str, bool], None] | None = None
        self._on_light_status: Callable[[str, LightStatus], None] | None = None
        self._on_motion_status: Callable[[str, MotionStatus], None] | None = None
        self._on_illuminance_status: Callable[[str, IlluminanceStatus], None] | None = (
            None
        )
        self._on_panel_status: Callable[[str, PanelStatus], None] | None = None
        self._on_energy_report: Callable[[str, float], None] | None = None
        self._on_sensor_on_off: Callable[[str, bool], None] | None = None
        self._on_energy: Callable[[str, Dict[str, Any]], None] | None = None

        self._window_ms = 100
        self._pending_requests: Dict[str, Dict[str, Dict[str, Any]]] = {}
        self._batch_timer: Dict[str, asyncio.TimerHandle] = {}  # cmd -> timer

    def _get_device_key(self, dev_type: str, channel: int, address: int) -> str:
        return f"{dev_type}_{channel}_{address}"

    def add_request(
        self, cmd: str, dev_type: str, channel: int, address: int, data: Dict[str, Any]
    ) -> None:
        if cmd not in self._pending_requests:
            self._pending_requests[cmd] = {}

        device_key = self._get_device_key(dev_type, channel, address)

        # Merge properties instead of overwriting the entire data
        if device_key in self._pending_requests[cmd]:
            existing_data = self._pending_requests[cmd][device_key]
            if "property" in existing_data and "property" in data:
                # Merge properties, avoiding duplicates by dpid
                existing_properties = {
                    prop["dpid"]: prop for prop in existing_data["property"]
                }
                new_properties = {prop["dpid"]: prop for prop in data["property"]}
                existing_properties.update(new_properties)
                data["property"] = list(existing_properties.values())

        self._pending_requests[cmd][device_key] = data

        if self._batch_timer.get(cmd) is None:
            self._batch_timer[cmd] = asyncio.get_event_loop().call_later(
                self._window_ms / 1000.0, self._flush_batch, cmd
            )

    def _flush_batch(self, cmd: str) -> None:
        if not self._pending_requests.get(cmd):
            return

        batch_data: List[Dict[str, Any]] = list(self._pending_requests[cmd].values())

        command: Dict[str, Any] = {
            "cmd": cmd,
            "msgId": str(int(time.time())),
            "gwSn": self._gw_sn,
            "data": batch_data,
        }

        self._mqtt_client.publish(self._pub_topic, json.dumps(command))

        _LOGGER.debug(
            "Gateway %s: Sent batch %s %s", self._gw_sn, cmd, json.dumps(command)
        )

        self._pending_requests[cmd].clear()
        self._batch_timer.pop(cmd)

    def __repr__(self) -> str:
        return (
            f"DaliGateway(gw_sn={self._gw_sn}, gw_ip={self._gw_ip}, "
            f"port={self._port}, name={self._name})"
        )

    @property
    def gw_sn(self) -> str:
        return self._gw_sn

    @property
    def gw_ip(self) -> str:
        return self._gw_ip

    @property
    def port(self) -> int:
        return self._port

    @property
    def username(self) -> str:
        return self._username

    @property
    def passwd(self) -> str:
        return self._passwd

    @property
    def channel_total(self) -> List[int]:
        return list(self._channel_total)

    @property
    def is_tls(self) -> bool:
        return self._is_tls

    @property
    def name(self) -> str:
        return self._name

    @property
    def on_online_status(self) -> Callable[[str, bool], None] | None:
        return self._on_online_status

    @on_online_status.setter
    def on_online_status(self, callback: Callable[[str, bool], None]) -> None:
        self._on_online_status = callback

    @property
    def on_light_status(self) -> Callable[[str, LightStatus], None] | None:
        return self._on_light_status

    @on_light_status.setter
    def on_light_status(self, callback: Callable[[str, LightStatus], None]) -> None:
        self._on_light_status = callback

    @property
    def on_motion_status(self) -> Callable[[str, MotionStatus], None] | None:
        return self._on_motion_status

    @on_motion_status.setter
    def on_motion_status(self, callback: Callable[[str, MotionStatus], None]) -> None:
        self._on_motion_status = callback

    @property
    def on_illuminance_status(
        self,
    ) -> Callable[[str, IlluminanceStatus], None] | None:
        return self._on_illuminance_status

    @on_illuminance_status.setter
    def on_illuminance_status(
        self, callback: Callable[[str, IlluminanceStatus], None]
    ) -> None:
        self._on_illuminance_status = callback

    @property
    def on_panel_status(self) -> Callable[[str, PanelStatus], None] | None:
        return self._on_panel_status

    @on_panel_status.setter
    def on_panel_status(self, callback: Callable[[str, PanelStatus], None]) -> None:
        self._on_panel_status = callback

    @property
    def on_energy_report(self) -> Callable[[str, float], None] | None:
        return self._on_energy_report

    @on_energy_report.setter
    def on_energy_report(self, callback: Callable[[str, float], None]) -> None:
        self._on_energy_report = callback

    @property
    def on_sensor_on_off(self) -> Callable[[str, bool], None] | None:
        return self._on_sensor_on_off

    @on_sensor_on_off.setter
    def on_sensor_on_off(self, callback: Callable[[str, bool], None]) -> None:
        self._on_sensor_on_off = callback

    @property
    def on_energy(self) -> Callable[[str, Dict[str, Any]], None] | None:
        return self._on_energy

    @on_energy.setter
    def on_energy(self, callback: Callable[[str, Dict[str, Any]], None]) -> None:
        self._on_energy = callback

    def _on_connect(
        self,
        client: paho_mqtt.Client,
        userdata: Any,
        flags: Any,
        rc: int,
        properties: Any = None,
    ) -> None:
        self._connect_result = rc
        self._connection_event.set()

        if rc == 0:
            _LOGGER.debug(
                "Gateway %s: MQTT connection established to %s:%s",
                self._gw_sn,
                self._gw_ip,
                self._port,
            )
            self._mqtt_client.subscribe(self._sub_topic)
            _LOGGER.debug(
                "Gateway %s: Subscribed to MQTT topic %s", self._gw_sn, self._sub_topic
            )

            # Trigger online_status callback with gateway SN as device ID and True status
            if self._on_online_status:
                self._on_online_status(self._gw_sn, True)
        else:
            _LOGGER.error(
                "Gateway %s: MQTT connection failed with code %s", self._gw_sn, rc
            )

    def _on_disconnect(
        self,
        client: paho_mqtt.Client,
        userdata: Any,
        *args: Any,
    ) -> None:
        # Handle different paho-mqtt versions:
        # v1.6.x: (client, userdata, rc)
        # v2.0.0+: (client, userdata, disconnect_flags, reason_code, properties)
        if HAS_CALLBACK_API_VERSION and len(args) >= 2:
            # paho-mqtt >= 2.0.0
            reason_code = args[1]  # disconnect_flags, reason_code, properties
        elif len(args) >= 1:
            # paho-mqtt < 2.0.0
            reason_code = args[0]  # rc
        else:
            reason_code = 0

        if reason_code != 0:
            _LOGGER.warning(
                "Gateway %s: Unexpected MQTT disconnection (%s:%s) - Reason code: %s",
                self._gw_sn,
                self._gw_ip,
                self._port,
                reason_code,
            )
        else:
            _LOGGER.debug("Gateway %s: MQTT disconnection completed", self._gw_sn)

        # Trigger online_status callback with gateway SN as device ID and False status
        if self._on_online_status:
            self._on_online_status(self._gw_sn, False)

    def _on_message(
        self, client: paho_mqtt.Client, userdata: Any, msg: paho_mqtt.MQTTMessage
    ) -> None:
        try:
            payload_json = json.loads(msg.payload.decode("utf-8", errors="replace"))
            _LOGGER.debug(
                "Gateway %s: Received MQTT message on topic %s: %s",
                self._gw_sn,
                msg.topic,
                payload_json,
            )

            cmd = payload_json.get("cmd")
            if not cmd:
                _LOGGER.warning(
                    "Gateway %s: Received MQTT message without cmd field", self._gw_sn
                )
                return

            command_handlers: Dict[str, Callable[[Dict[str, Any]], None]] = {
                "devStatus": self._process_device_status,
                "readDevRes": self._process_device_status,
                "writeDevRes": self._process_write_response,
                "writeGroupRes": self._process_write_response,
                "writeSceneRes": self._process_write_response,
                "onlineStatus": self._process_online_status,
                "reportEnergy": self._process_energy_report,
                "searchDevRes": self._process_search_device_response,
                "getSceneRes": self._process_get_scene_response,
                "getGroupRes": self._process_get_group_response,
                "getVersionRes": self._process_get_version_response,
                "readGroupRes": self._process_read_group_response,
                "readSceneRes": self._process_read_scene_response,
                "restartGatewayRes": self._process_restart_gateway_response,
                "getEnergyRes": self._process_get_energy_response,
                "setSensorOnOffRes": self._process_set_sensor_on_off_response,
                "getSensorOnOffRes": self._process_get_sensor_on_off_response,
                "setDevParamRes": self._process_write_response,
                "getDevParamRes": self._process_get_dev_param_response,
            }

            handler = command_handlers.get(cmd)
            if handler:
                handler(payload_json)
            else:
                _LOGGER.debug(
                    "Gateway %s: Unhandled MQTT command '%s', payload: %s",
                    self._gw_sn,
                    cmd,
                    payload_json,
                )

        except json.JSONDecodeError:
            _LOGGER.error(
                "Gateway %s: Failed to decode MQTT message payload: %s",
                self._gw_sn,
                msg.payload,
            )
        except (ValueError, KeyError, TypeError) as e:
            _LOGGER.error(
                "Gateway %s: Error processing MQTT message: %s", self._gw_sn, str(e)
            )

    def _process_online_status(self, payload: Dict[str, Any]) -> None:
        data_list = payload.get("data")
        if not data_list:
            _LOGGER.warning(
                "Gateway %s: Received onlineStatus with no data: %s",
                self._gw_sn,
                payload,
            )
            return

        for data in data_list:
            dev_id = gen_device_unique_id(
                data.get("devType"),
                data.get("channel"),
                data.get("address"),
                self._gw_sn,
            )

            available: bool = data.get("status", False)

            if self._on_online_status:
                self._on_online_status(dev_id, available)

    def _process_device_status(self, payload: Dict[str, Any]) -> None:
        data = payload.get("data")
        if not data:
            _LOGGER.warning(
                "Gateway %s: Received devStatus with no data: %s", self._gw_sn, payload
            )
            return

        dev_id = gen_device_unique_id(
            data.get("devType"), data.get("channel"), data.get("address"), self._gw_sn
        )

        if not dev_id:
            _LOGGER.warning("Failed to generate device ID from data: %s", data)
            return

        property_list = data.get("property", [])
        dev_type = data.get("devType")

        if dev_type and is_light_device(dev_type) and self._on_light_status:
            light_status = parse_light_status(property_list)
            self._on_light_status(dev_id, light_status)
        elif dev_type and is_motion_sensor(dev_type) and self._on_motion_status:
            motion_statuses = parse_motion_status(property_list)
            for motion_status in motion_statuses:
                self._on_motion_status(dev_id, motion_status)
        elif (
            dev_type and is_illuminance_sensor(dev_type) and self._on_illuminance_status
        ):
            illuminance_statuses = parse_illuminance_status(property_list)
            for illuminance_status in illuminance_statuses:
                self._on_illuminance_status(dev_id, illuminance_status)
        elif dev_type and is_panel_device(dev_type) and self._on_panel_status:
            panel_statuses = parse_panel_status(property_list)
            for panel_status in panel_statuses:
                self._on_panel_status(dev_id, panel_status)
        else:
            # Warn if no callback handler exists for this device type
            _LOGGER.warning(
                "Gateway %s: No callback handler for device type %s (device: %s). "
                "Property data: %s",
                self._gw_sn,
                dev_type,
                dev_id,
                property_list,
            )

    def _process_write_response(self, payload: Dict[str, Any]) -> None:
        msg_id = payload.get("msgId")
        ack = payload.get("ack", False)

        _LOGGER.debug(
            "Gateway %s: Received write device response, "
            "msgId: %s, ack: %s, payload: %s",
            self._gw_sn,
            msg_id,
            ack,
            payload,
        )

    def _process_energy_report(self, payload: Dict[str, Any]) -> None:
        data = payload.get("data")
        if not data:
            _LOGGER.warning(
                "Gateway %s: Received reportEnergy with no data: %s",
                self._gw_sn,
                payload,
            )
            return

        dev_id = gen_device_unique_id(
            data.get("devType"), data.get("channel"), data.get("address"), self._gw_sn
        )

        if not dev_id:
            _LOGGER.warning("Failed to generate device ID from data: %s", data)
            return

        property_list = data.get("property", [])
        for prop in property_list:
            if prop.get("dpid") == 30:
                try:
                    energy_value = float(prop.get("value", "0"))

                    if self._on_energy_report:
                        self._on_energy_report(dev_id, energy_value)
                except (ValueError, TypeError) as e:
                    _LOGGER.error("Error converting energy value: %s", str(e))

    def _process_get_version_response(self, payload_json: Dict[str, Any]) -> None:
        self._version_result = VersionType(
            software=payload_json.get("data", {}).get("swVersion", ""),
            firmware=payload_json.get("data", {}).get("fwVersion", ""),
        )
        self._version_received.set()

    def _process_get_energy_response(self, payload_json: Dict[str, Any]) -> None:
        data_list = payload_json.get("data")
        if not data_list:
            _LOGGER.warning(
                "Gateway %s: Received getEnergyRes with no data: %s",
                self._gw_sn,
                payload_json,
            )
            return

        for data in data_list:
            dev_id = gen_device_unique_id(
                data.get("devType"),
                data.get("channel"),
                data.get("address"),
                self._gw_sn,
            )

            if not dev_id:
                _LOGGER.warning("Failed to generate device ID from data: %s", data)
                continue

            energy_data = {
                "yearEnergy": data.get("yearEnergy", {}),
                "monthEnergy": data.get("monthEnergy", {}),
                "dayEnergy": data.get("dayEnergy", {}),
                "hourEnergy": data.get("hourEnergy", []),
            }

            if self._on_energy:
                self._on_energy(dev_id, energy_data)

    def _process_search_device_response(self, payload_json: Dict[str, Any]) -> None:
        for raw_device_data in payload_json.get("data", []):
            dev_type = str(raw_device_data.get("devType", ""))
            channel = int(raw_device_data.get("channel", 0))
            address = int(raw_device_data.get("address", 0))

            unique_id = gen_device_unique_id(dev_type, channel, address, self._gw_sn)
            dev_id = str(raw_device_data.get("devId") or unique_id)
            name = str(
                raw_device_data.get("name")
                or gen_device_name(dev_type, channel, address)
            )

            device = Device(
                self,
                unique_id=unique_id,
                dev_id=dev_id,
                name=name,
                dev_type=dev_type,
                channel=channel,
                address=address,
                status=str(raw_device_data.get("status", "")),
                dev_sn=str(raw_device_data.get("devSn", "")),
                area_name=str(raw_device_data.get("areaName", "")),
                area_id=str(raw_device_data.get("areaId", "")),
                model=DEVICE_MODEL_MAP.get(dev_type, "Unknown"),
                properties=[],
            )

            if not any(
                existing.unique_id == device.unique_id
                for existing in self._devices_result
            ):
                self._devices_result.append(device)

        search_status = payload_json.get("searchStatus")
        if search_status in {0, 1}:
            self._devices_received.set()

    def _process_get_scene_response(self, payload_json: Dict[str, Any]) -> None:
        self._scenes_result.clear()
        for channel_scenes in payload_json.get("scene", []):
            channel = channel_scenes.get("channel", 0)

            for scene_data in channel_scenes.get("data", []):
                scene_id = int(scene_data.get("sceneId", 0))
                name = str(scene_data.get("name", ""))
                area_id = str(scene_data.get("areaId", ""))

                if any(
                    existing.unique_id
                    == gen_scene_unique_id(scene_id, channel, self._gw_sn)
                    for existing in self._scenes_result
                ):
                    continue

                self._scenes_result.append(
                    Scene(
                        self,
                        scene_id=scene_id,
                        name=name,
                        channel=channel,
                        area_id=area_id,
                    )
                )

        self._scenes_received.set()

    def _process_get_group_response(self, payload_json: Dict[str, Any]) -> None:
        self._groups_result.clear()
        for channel_groups in payload_json.get("group", []):
            channel = channel_groups.get("channel", 0)

            for group_data in channel_groups.get("data", []):
                group_id = int(group_data.get("groupId", 0))
                name = str(group_data.get("name", ""))
                area_id = str(group_data.get("areaId", ""))

                if any(
                    existing.unique_id
                    == gen_group_unique_id(group_id, channel, self._gw_sn)
                    for existing in self._groups_result
                ):
                    continue

                self._groups_result.append(
                    Group(
                        self,
                        group_id=group_id,
                        name=name,
                        channel=channel,
                        area_id=area_id,
                    )
                )

        self._groups_received.set()

    def _process_read_group_response(self, payload: Dict[str, Any]) -> None:
        group_id = payload.get("groupId", 0)
        group_name = payload.get("name", "")
        channel = payload.get("channel", 0)
        raw_devices = payload.get("data", [])

        devices: List[Dict[str, Any]] = []
        for device_data in raw_devices:
            dev_type = str(device_data.get("devType", ""))
            channel_id = int(device_data.get("channel", 0))
            address = int(device_data.get("address", 0))

            devices.append(
                {
                    "unique_id": gen_device_unique_id(
                        dev_type, channel_id, address, self._gw_sn
                    ),
                    "id": str(device_data.get("devId", "")),
                    "name": gen_device_name(dev_type, channel_id, address),
                    "dev_type": dev_type,
                    "channel": channel_id,
                    "address": address,
                    "status": "",
                    "dev_sn": "",
                    "area_name": "",
                    "area_id": "",
                    "model": DEVICE_MODEL_MAP.get(dev_type, "Unknown"),
                    "prop": [],
                }
            )

        self._read_group_result = {
            "unique_id": gen_group_unique_id(group_id, channel, self._gw_sn),
            "id": group_id,
            "name": group_name,
            "channel": channel,
            "area_id": "",
            "devices": devices,
        }
        self._read_group_received.set()

    def _process_read_scene_response(self, payload: Dict[str, Any]) -> None:
        scene_id = payload.get("sceneId", 0)
        scene_name = payload.get("name", "")
        channel = payload.get("channel", 0)
        data: Dict[str, Any] | None = payload.get("data")

        if data is None:
            _LOGGER.error(
                "Gateway %s: Received readSceneRes with no data: %s",
                self._gw_sn,
                payload,
            )
            self._read_scene_received.set()
            return

        raw_devices: List[Dict[str, Any]] = data.get("device", [])

        # Create SceneDeviceType objects from raw device data
        devices: List[SceneDeviceType] = []
        for device_data in raw_devices:
            # Convert raw property data to LightStatus using parse_light_status
            raw_properties = device_data.get("property", [])
            light_status = parse_light_status(raw_properties)

            device: SceneDeviceType = {
                "dev_type": device_data.get("devType", ""),
                "channel": device_data.get("channel", 0),
                "address": device_data.get("address", 0),
                "gw_sn_obj": device_data.get("gwSnObj", ""),
                "property": light_status,
            }
            devices.append(device)

        self._read_scene_result = {
            "unique_id": gen_scene_unique_id(scene_id, channel, self._gw_sn),
            "id": scene_id,
            "name": scene_name,
            "channel": channel,
            "area_id": "",
            "devices": devices,
        }
        self._read_scene_received.set()

    def _process_set_sensor_on_off_response(self, payload: Dict[str, Any]) -> None:
        _LOGGER.debug(
            "Gateway %s: Received setSensorOnOffRes response, payload: %s",
            self._gw_sn,
            payload,
        )

    def _process_get_sensor_on_off_response(self, payload: Dict[str, Any]) -> None:
        dev_id = gen_device_unique_id(
            payload.get("devType", ""),
            payload.get("channel", 0),
            payload.get("address", 0),
            self._gw_sn,
        )

        value = payload.get("value", False)

        if self._on_sensor_on_off:
            self._on_sensor_on_off(dev_id, value)

    def _process_get_dev_param_response(self, payload: Dict[str, Any]) -> None:
        _LOGGER.debug(
            "Gateway %s: Received getDevParamRes response, payload: %s",
            self._gw_sn,
            payload,
        )

    def _process_restart_gateway_response(self, payload: Dict[str, Any]) -> None:
        ack = payload.get("ack", False)
        _LOGGER.info(
            "Gateway %s: Received restart confirmation, ack: %s. Gateway will restart shortly.",
            self._gw_sn,
            ack,
        )

    async def _setup_ssl(self) -> None:
        try:
            loop = asyncio.get_event_loop()
            await loop.run_in_executor(None, self._setup_ssl_sync)
        except Exception as e:
            _LOGGER.error("Failed to configure SSL/TLS: %s", str(e))
            raise DaliGatewayError(
                f"SSL/TLS configuration failed: {e}", self._gw_sn
            ) from e

    def _setup_ssl_sync(self) -> None:
        context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
        context.load_verify_locations(str(CA_CERT_PATH))
        context.check_hostname = False
        context.verify_mode = ssl.CERT_REQUIRED
        self._mqtt_client.tls_set_context(context)
        _LOGGER.debug("SSL/TLS configured with CA certificate: %s", CA_CERT_PATH)

    def get_credentials(self) -> tuple[str, str]:
        return self._username, self._passwd

    async def connect(self) -> None:
        self._connection_event.clear()
        self._connect_result = None
        self._mqtt_client.username_pw_set(self._username, self._passwd)

        if self._is_tls:
            await self._setup_ssl()

        try:
            _LOGGER.info(
                "Attempting connection to gateway %s at %s:%s (TLS: %s)",
                self._gw_sn,
                self._gw_ip,
                self._port,
                self._is_tls,
            )
            self._mqtt_client.connect(self._gw_ip, self._port)
            self._mqtt_client.loop_start()
            await asyncio.wait_for(self._connection_event.wait(), timeout=10)

            if self._connect_result is not None and self._connect_result == 0:
                _LOGGER.info(
                    "Successfully connected to gateway %s at %s:%s",
                    self._gw_sn,
                    self._gw_ip,
                    self._port,
                )
                return

        except asyncio.TimeoutError as err:
            _LOGGER.error(
                "Connection timeout to gateway %s at %s:%s after 10 seconds - check network connectivity",
                self._gw_sn,
                self._gw_ip,
                self._port,
            )
            raise DaliGatewayError(
                f"Connection timeout to gateway {self._gw_sn}", self._gw_sn
            ) from err
        except (ConnectionRefusedError, OSError) as err:
            _LOGGER.error(
                "Network error connecting to gateway %s at %s:%s: %s - check if gateway is powered on and accessible",
                self._gw_sn,
                self._gw_ip,
                self._port,
                str(err),
            )
            raise DaliGatewayError(
                f"Network error connecting to gateway {self._gw_sn}: {err}", self._gw_sn
            ) from err

        if self._connect_result is not None and self._connect_result in (4, 5):
            _LOGGER.error(
                "Authentication failed for gateway %s (code %s) with credentials user='%s'. "
                "Please press the gateway button and retry",
                self._gw_sn,
                self._connect_result,
                self._username,
            )
            raise DaliGatewayError(
                f"Authentication failed for gateway {self._gw_sn}. "
                "Please press the gateway button and retry",
                self._gw_sn,
            )
        _LOGGER.error(
            "Connection failed for gateway %s with result code %s",
            self._gw_sn,
            self._connect_result,
        )
        raise DaliGatewayError(
            f"Connection failed for gateway {self._gw_sn} "
            f"with code {self._connect_result}"
        )

    async def disconnect(self) -> None:
        try:
            self._mqtt_client.loop_stop()
            self._mqtt_client.disconnect()
            self._connection_event.clear()
            _LOGGER.info("Successfully disconnected from gateway %s", self._gw_sn)
        except Exception as exc:  # pylint: disable=broad-exception-caught
            _LOGGER.error(
                "Error during disconnect from gateway %s: %s", self._gw_sn, exc
            )
            raise DaliGatewayError(
                f"Failed to disconnect from gateway {self._gw_sn}: {exc}"
            ) from exc

    async def get_version(self) -> VersionType | None:
        self._version_received = asyncio.Event()
        payload = {
            "cmd": "getVersion",
            "msgId": str(int(time.time())),
            "gwSn": self._gw_sn,
        }

        _LOGGER.debug("Gateway %s: Sending get version command", self._gw_sn)
        self._mqtt_client.publish(self._pub_topic, json.dumps(payload))

        try:
            await asyncio.wait_for(self._version_received.wait(), timeout=30.0)
        except asyncio.TimeoutError:
            _LOGGER.warning(
                "Gateway %s: Timeout waiting for version response", self._gw_sn
            )

        _LOGGER.info(
            "Gateway %s: Version retrieved - SW: %s, FW: %s",
            self._gw_sn,
            self._version_result["software"] if self._version_result else "N/A",
            self._version_result["firmware"] if self._version_result else "N/A",
        )
        return self._version_result

    async def read_group(self, group_id: int, channel: int = 0) -> Dict[str, Any]:
        self._read_group_received = asyncio.Event()
        payload: Dict[str, Any] = {
            "cmd": "readGroup",
            "msgId": str(int(time.time())),
            "gwSn": self._gw_sn,
            "channel": channel,
            "groupId": group_id,
        }

        _LOGGER.debug("Gateway %s: Sending read group command", self._gw_sn)
        self._mqtt_client.publish(self._pub_topic, json.dumps(payload))

        try:
            await asyncio.wait_for(self._read_group_received.wait(), timeout=30.0)
        except asyncio.TimeoutError as err:
            _LOGGER.error(
                "Gateway %s: Timeout waiting for read group response for group %s",
                self._gw_sn,
                group_id,
            )
            raise DaliGatewayError(
                f"Timeout reading group {group_id} from gateway {self._gw_sn}",
                self._gw_sn,
            ) from err

        if not self._read_group_result:
            _LOGGER.error(
                "Gateway %s: Failed to read group %s - group may not exist",
                self._gw_sn,
                group_id,
            )
            raise DaliGatewayError(
                f"Failed to read group {group_id} from gateway {self._gw_sn}. Group may not exist.",
                self._gw_sn,
            )

        _LOGGER.info(
            "Gateway %s: Group read completed - ID: %s, Name: %s, Devices: %d",
            self._gw_sn,
            self._read_group_result["id"],
            self._read_group_result["name"],
            len(self._read_group_result["devices"]),
        )
        return self._read_group_result

    async def read_scene(self, scene_id: int, channel: int = 0) -> Dict[str, Any]:
        self._read_scene_received = asyncio.Event()
        payload: Dict[str, Any] = {
            "cmd": "readScene",
            "msgId": str(int(time.time())),
            "gwSn": self._gw_sn,
            "channel": channel,
            "sceneId": scene_id,
        }

        _LOGGER.debug("Gateway %s: Sending read scene command", self._gw_sn)
        self._mqtt_client.publish(self._pub_topic, json.dumps(payload))

        try:
            await asyncio.wait_for(self._read_scene_received.wait(), timeout=30.0)
        except asyncio.TimeoutError as err:
            _LOGGER.error(
                "Gateway %s: Timeout waiting for read scene response for scene %s",
                self._gw_sn,
                scene_id,
            )
            raise DaliGatewayError(
                f"Timeout reading scene {scene_id} from gateway {self._gw_sn}",
                self._gw_sn,
            ) from err

        if not self._read_scene_result:
            _LOGGER.error(
                "Gateway %s: Failed to read scene %s - scene may not exist",
                self._gw_sn,
                scene_id,
            )
            raise DaliGatewayError(
                f"Scene {scene_id} not found on gateway {self._gw_sn}", self._gw_sn
            )

        _LOGGER.info(
            "Gateway %s: Scene read completed - ID: %s, Name: %s, Devices: %d",
            self._gw_sn,
            self._read_scene_result["id"],
            self._read_scene_result["name"],
            len(self._read_scene_result["devices"]),
        )
        return self._read_scene_result

    async def discover_devices(self) -> list[Device]:
        self._devices_received = asyncio.Event()
        self._devices_result.clear()
        search_payload = {
            "cmd": "searchDev",
            "searchFlag": "exited",
            "msgId": str(int(time.time())),
            "gwSn": self._gw_sn,
        }

        _LOGGER.debug("Gateway %s: Sending device discovery command", self._gw_sn)
        self._mqtt_client.publish(self._pub_topic, json.dumps(search_payload))

        try:
            await asyncio.wait_for(self._devices_received.wait(), timeout=30.0)
        except asyncio.TimeoutError:
            _LOGGER.warning(
                "Gateway %s: Timeout waiting for device discovery response", self._gw_sn
            )

        _LOGGER.info(
            "Gateway %s: Device discovery completed, found %d device(s)",
            self._gw_sn,
            len(self._devices_result),
        )
        return self._devices_result

    async def discover_groups(self) -> list[Group]:
        self._groups_received = asyncio.Event()
        self._groups_result.clear()
        search_payload = {
            "cmd": "getGroup",
            "msgId": str(int(time.time())),
            "getFlag": "exited",
            "gwSn": self._gw_sn,
        }

        _LOGGER.debug("Gateway %s: Sending group discovery command", self._gw_sn)
        self._mqtt_client.publish(self._pub_topic, json.dumps(search_payload))

        try:
            await asyncio.wait_for(self._groups_received.wait(), timeout=30.0)
        except asyncio.TimeoutError:
            _LOGGER.warning(
                "Gateway %s: Timeout waiting for group discovery response", self._gw_sn
            )

        _LOGGER.info(
            "Gateway %s: Group discovery completed, found %d group(s)",
            self._gw_sn,
            len(self._groups_result),
        )
        return self._groups_result

    async def discover_scenes(self) -> list[Scene]:
        self._scenes_received = asyncio.Event()
        self._scenes_result.clear()
        search_payload = {
            "cmd": "getScene",
            "msgId": str(int(time.time())),
            "getFlag": "exited",
            "gwSn": self._gw_sn,
        }

        _LOGGER.debug("Gateway %s: Sending scene discovery command", self._gw_sn)
        self._mqtt_client.publish(self._pub_topic, json.dumps(search_payload))

        try:
            await asyncio.wait_for(self._scenes_received.wait(), timeout=30.0)
        except asyncio.TimeoutError:
            _LOGGER.warning(
                "Gateway %s: Timeout waiting for scene discovery response", self._gw_sn
            )

        _LOGGER.info(
            "Gateway %s: Scene discovery completed, found %d scene(s)",
            self._gw_sn,
            len(self._scenes_result),
        )
        return self._scenes_result

    def command_write_dev(
        self,
        dev_type: str,
        channel: int,
        address: int,
        properties: List[Dict[str, Any]],
    ) -> None:
        self.add_request(
            "writeDev",
            dev_type,
            channel,
            address,
            {
                "devType": dev_type,
                "channel": channel,
                "address": address,
                "property": properties,
            },
        )

    def command_read_dev(self, dev_type: str, channel: int, address: int) -> None:
        self.add_request(
            "readDev",
            dev_type,
            channel,
            address,
            {"devType": dev_type, "channel": channel, "address": address},
        )

    def command_get_energy(
        self, dev_type: str, channel: int, address: int, year: int, month: int, day: int
    ) -> None:
        self.add_request(
            "getEnergy",
            dev_type,
            channel,
            address,
            {
                "devType": dev_type,
                "channel": channel,
                "address": address,
                "condition": {"year": year, "month": month, "day": day, "hour": []},
            },
        )

    def command_write_group(
        self, group_id: int, channel: int, properties: List[Dict[str, Any]]
    ) -> None:
        command: Dict[str, Any] = {
            "cmd": "writeGroup",
            "msgId": str(int(time.time())),
            "gwSn": self._gw_sn,
            "channel": channel,
            "groupId": group_id,
            "data": properties,
        }
        command_json = json.dumps(command)
        self._mqtt_client.publish(self._pub_topic, command_json)

    def command_write_scene(self, scene_id: int, channel: int) -> None:
        command: Dict[str, Any] = {
            "cmd": "writeScene",
            "msgId": str(int(time.time())),
            "gwSn": self._gw_sn,
            "channel": channel,
            "sceneId": scene_id,
        }
        command_json = json.dumps(command)
        self._mqtt_client.publish(self._pub_topic, command_json)

    def command_set_sensor_on_off(
        self, dev_type: str, channel: int, address: int, value: bool
    ) -> None:
        command: Dict[str, Any] = {
            "cmd": "setSensorOnOff",
            "msgId": str(int(time.time())),
            "gwSn": self._gw_sn,
            "devType": dev_type,
            "channel": channel,
            "address": address,
            "value": value,
        }
        command_json = json.dumps(command)
        self._mqtt_client.publish(self._pub_topic, command_json)

    def command_get_sensor_on_off(
        self, dev_type: str, channel: int, address: int
    ) -> None:
        command: Dict[str, Any] = {
            "cmd": "getSensorOnOff",
            "msgId": str(int(time.time())),
            "gwSn": self._gw_sn,
            "devType": dev_type,
            "channel": channel,
            "address": address,
        }
        command_json = json.dumps(command)
        self._mqtt_client.publish(self._pub_topic, command_json)

    def command_get_dev_param(self, dev_type: str, channel: int, address: int) -> None:
        command: Dict[str, Any] = {
            "cmd": "getDevParam",
            "msgId": str(int(time.time())),
            "gwSn": self._gw_sn,
            "devType": dev_type,
            "channel": channel,
            "address": address,
            "fromBus": True,
        }
        command_json = json.dumps(command)
        self._mqtt_client.publish(self._pub_topic, command_json)

    def command_set_dev_param(
        self, dev_type: str, channel: int, address: int, param: DeviceParamType
    ) -> None:
        command: Dict[str, Any] = {
            "cmd": "setDevParam",
            "msgId": str(int(time.time())),
            "gwSn": self._gw_sn,
            "data": [
                {
                    "devType": dev_type,
                    "channel": channel,
                    "address": address,
                    "paramer": {
                        "maxBrightness": param["max_brightness"],
                    },
                }
            ],
        }
        command_json = json.dumps(command)
        _LOGGER.debug(
            "Gateway %s: Sending setDevParam command: %s", self._gw_sn, command
        )
        self._mqtt_client.publish(self._pub_topic, command_json)

    def restart_gateway(self) -> None:
        """Restart the gateway."""
        command: Dict[str, Any] = {
            "cmd": "restartGateway",
            "msgId": str(int(time.time())),
        }
        command_json = json.dumps(command)
        _LOGGER.debug("Gateway %s: Sending restart command", self._gw_sn)
        self._mqtt_client.publish(self._pub_topic, command_json)
